@@ -10,7 +10,7 @@ source "$TEST_TOP/common"
 check_prereq mkfs.btrfs
 prepare_test_dev
 
-tmp=$(mktemp -d --tmpdir btrfs-progs-mkfs.rootdirXXXXXXX)
+tmp=$(_mktemp_dir mkfs-rootdir)
 
 fallocate -l 128M $tmp/large_file
 
